name = 'resource_man'
version = '2.1.12'
description = 'Python resource manager to work with the standard library for regular python code and executables.'
url = 'https://github.com/justengel/resource_man'
author = 'Justin Engel'
author_email = 'jtengel08@gmail.com'
